#ifdef HAVE_CONFIG_H
#include <linux/config.h>
#endif

#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/rwsem.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/workqueue.h>
#include <linux/kthread.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a<<16)|(b<<8)|c)
#endif

#if LINUX_VERSION_CODE<KERNEL_VERSION(2,6,27)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif

#include "wrg3ioctl.h"

#define MAX_WRG3_DEVICES 16
#define WiNRADiO_MAJOR     82
#define WRG3_MINOR_BASE  32
#define WRG3_MINOR_END   (WRG3_MINOR_BASE+MAX_WRG3_DEVICES-1)

#define WRG3_NAME "wrg313"

#define WRG3_PCI 0
#define WRG3_USB 1

#define WRG3_DISPLAY_NAME "WRG313"

#define MAX_PIC_RETRY 128

#define USB_TIMEOUT (2000)

#define MAX_ENDPOINTS 16

#define MAX_BULK_URBS 8

#define TO_JIFFIES(x) ((x*HZ)/1000)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
#define USE_MUTEX
#else
#undef USE_MUTEX
#endif

#ifdef USE_MUTEX

#include <linux/mutex.h>

#define WRG3_MUTEX struct mutex
#define WRG3_INIT_MUTEX(object)         mutex_init(&object)
#define WRG3_DESTROY_MUTEX(object)      mutex_destroy(&object)
#define WRG3_LOCK(object)               mutex_lock(&object)
#define WRG3_LOCK_INTERRUPTIBLE(object) mutex_lock_interruptible(&object)
#define WRG3_UNLOCK(object)             mutex_unlock(&object)

#else

#define WRG3_MUTEX struct semaphore
#define WRG3_INIT_MUTEX(object)         init_MUTEX(&object)
#define WRG3_DESTROY_MUTEX(object)
#define WRG3_LOCK(object)               down(&object)
#define WRG3_LOCK_INTERRUPTIBLE(object) down_interruptible(&object)
#define WRG3_UNLOCK(object)             up(&object)

#endif

typedef struct _USB_BUFFER
{
    __u8    *Buffer;
    __u32   Used;
    __u32   Remain;
    struct _USB_BUFFER *Next;
} USB_BUFFER,*PUSB_BUFFER;


typedef struct _ISOC_URB
{
    struct urb               *Urb;
    struct __WRG3          *Dev;
    struct _ISOC_URB         *Next;
    __u8                     *Buffer;
} ISOC_URB,*PISOC_URB;


typedef struct __WRG3
{
    char Name[32];

    int DeviceIndex;
    int Minor;

    struct device Device;

    struct cdev cdev;
    WRG3_MUTEX Ioctl_Lock;

    __u8 InterfaceType;
    __u8 Opened;
    __u8 Connected;

    union
    {
        struct //PCI device data
        {
            struct pci_dev          *PCIDevice;
            __u8                    *MemHPI;
            __u8                    *MemGP;
            int                     IrqNum;
            __u8                    IrqHandlerRegistered;
            struct semaphore        Interrupt_Semaphore;
            __u8                    InterruptStarted;
            spinlock_t              Interrupt_Lock;
        };

        struct //USB device data
        {
            struct usb_device       *USBDevice;
            struct usb_interface    *USBInterface;
            __u8                    NumberOfEndPoints;

            WRG3_MUTEX              Streaming_Lock;
            WRG3_MUTEX              Streaming_Buffers_Lock;
            PUSB_BUFFER             Streaming_Buffers;
            PUSB_BUFFER             Streaming_FirstBuffer;
            PUSB_BUFFER             Streaming_CurrentBuffer;
            __u32                   Streaming_BufferSize;
            __u32                   Streaming_Shift;
            struct completion       Streaming_Completion;
            __u8                    Streaming_Active;
            //struct task_struct      *Streaming_Thread;
            struct list_head        Streaming_CompleteIsocUrbs;
            struct semaphore        Streaming_CompleteIsocUrbs_Semaphore;
            spinlock_t              Streaming_CompleteIsocUrbs_Lock;
            PISOC_URB               Streaming_IsocUrbs;
            volatile int            Streaming_SentIsocUrbs;
            struct rw_semaphore     Streaming_Semaphore;

            spinlock_t              SentBulkUrbs_SpinLock;
            volatile int            SentBulkUrbs;

            struct semaphore        Streaming_Thread_Start_Semaphore;
            struct semaphore        Streaming_Thread_Stop_Semaphore;
            volatile int            Streaming_Thread_Stop;
            struct task_struct      *Streaming_Thread;
        };
    };
} WRG3,*PWRG3;

static PWRG3 WRG3_Devices[MAX_WRG3_DEVICES];

#ifdef USE_MUTEX
static DEFINE_MUTEX(WRG3_Device_Lock);
#else
static DECLARE_MUTEX(WRG3_Device_Lock);
#endif

static struct usb_device_id WRG3USB_ID_Table[]=
{
    {USB_DEVICE(0x14E0,0x3103)},
    {},
};

static struct pci_device_id WRG3PCI_ID_Table[]=
{
    {.vendor=0x104C,.device=0xAC60,.subvendor=0x186F,.subdevice=0x3101},
    {},
};

MODULE_DEVICE_TABLE(usb,WRG3USB_ID_Table);
MODULE_DEVICE_TABLE(pci,WRG3PCI_ID_Table);


int RegisterWRG3Device(PWRG3 Dev);
int UnregisterWRG3Device(PWRG3 Dev);
int WRG3PCI_StopInterrupt(PWRG3 Dev);


int WRG3_Release(struct inode *inode, struct file *filp);
int WRG3_Open(struct inode *inode, struct file *filp);

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
long WRG3_Ioctl(struct file *filp, unsigned int cmd, unsigned long arg);
#else
int WRG3_Ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
#endif

static struct file_operations WRG3_File_Operations=
{
    .open=WRG3_Open,
    .release=WRG3_Release,
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
    .unlocked_ioctl=WRG3_Ioctl,
#else
    .ioctl=WRG3_Ioctl,
#endif
};

static struct class WRG3_Class=
{
    .name=WRG3_NAME,
};

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
static void WRG3USB_IsoUrbComplete(struct urb *Urb);
#else
static void WRG3USB_IsoUrbComplete(struct urb *Urb,struct pt_regs *Regs);
#endif


PUSB_BUFFER WRG3USB_AllocateBuffer(int BufferSize)
{
 PUSB_BUFFER Buffer;

    Buffer=(PUSB_BUFFER)kmalloc(BufferSize+sizeof(USB_BUFFER),GFP_KERNEL);

    if(Buffer!=NULL)
    {
        Buffer->Buffer=(__u8*)(Buffer+1);
        Buffer->Used=0;
        Buffer->Remain=BufferSize;
        Buffer->Next=NULL;
    }

    return Buffer;
}

void WRG3USB_FreeStreamingBuffers(PWRG3 Dev)
{
 PUSB_BUFFER Buffer,Next;

    if(Dev->Streaming_Buffers!=NULL)
    {
        Buffer=Dev->Streaming_Buffers;

        do
        {
            Next=Buffer->Next;
            kfree(Buffer);
            Buffer=Next;
        }
        while(Buffer!=Dev->Streaming_Buffers);

        Dev->Streaming_Buffers=NULL;
    }
}

int WRG3USB_AllocateIsocUrb(PWRG3 Dev,__u8 EndPoint,__u32 BufferSize,PISOC_URB *Result)
{
 __u32 MaxPacketSize;
 __u32 NumberOfPackets,i;
 struct urb *Urb;
 __u8 *Buffer;
 __u32 Offset;
 PISOC_URB IsoUrb;

    if(BufferSize==0)
    {
        return -EINVAL;
    }

    if(EndPoint>15 || Dev->USBDevice->ep_in[EndPoint]==NULL)
    {
        return -EINVAL;
    }

    MaxPacketSize=Dev->USBDevice->ep_in[EndPoint]->desc.wMaxPacketSize;


    NumberOfPackets=(BufferSize+MaxPacketSize-1)/MaxPacketSize;

    Urb=usb_alloc_urb(NumberOfPackets,GFP_KERNEL);

    if(Urb==NULL)
    {
        return -ENOMEM;
    }


    Buffer=(__u8*)kmalloc(BufferSize,GFP_KERNEL);

    if(Buffer==NULL)
    {
        usb_free_urb(Urb);
        return -ENOMEM;
    }

    IsoUrb=(PISOC_URB)kmalloc(sizeof(ISOC_URB),GFP_KERNEL);

    if(IsoUrb==NULL)
    {
        usb_free_urb(Urb);
        kfree(Buffer);
        return -ENOMEM;
    }

    Urb->dev=Dev->USBDevice;
    Urb->pipe=usb_rcvisocpipe(Dev->USBDevice,EndPoint);
    Urb->interval=1;
    Urb->transfer_flags=URB_ISO_ASAP;
    Urb->transfer_buffer=Buffer;
    Urb->transfer_buffer_length=BufferSize;
    Urb->number_of_packets=NumberOfPackets;
    Urb->complete=WRG3USB_IsoUrbComplete;
    Urb->context=IsoUrb;

    Offset=0;



    for(i=0;i<NumberOfPackets;i++,Offset+=MaxPacketSize)
    {
        Urb->iso_frame_desc[i].offset=Offset;
        Urb->iso_frame_desc[i].length=BufferSize>=MaxPacketSize?MaxPacketSize:BufferSize;
        BufferSize-=MaxPacketSize;
    }

    IsoUrb->Urb=Urb;
    IsoUrb->Dev=Dev;
    IsoUrb->Next=NULL;
    IsoUrb->Buffer=Buffer;

    *Result=IsoUrb;

    return 0;
}

void WRG3USB_FreeIsocUrbs(PWRG3 Dev)
{
 PISOC_URB IsocUrb,Next;
 unsigned long Flags;

    if(Dev->Streaming_IsocUrbs!=NULL)
    {
        IsocUrb=Dev->Streaming_IsocUrbs;

        do
        {
            usb_kill_urb(IsocUrb->Urb);

            Next=IsocUrb->Next;

            usb_free_urb(IsocUrb->Urb);
            kfree(IsocUrb->Buffer);
            kfree(IsocUrb);

            IsocUrb=Next;
        }
        while(IsocUrb!=Dev->Streaming_IsocUrbs);

        Dev->Streaming_IsocUrbs=NULL;
    }

    spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

    Dev->Streaming_SentIsocUrbs=0;

    INIT_LIST_HEAD(&Dev->Streaming_CompleteIsocUrbs);

    spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);
}

void WRG3PCI_FreeResources(PWRG3 Dev)
{
    if(Dev->IrqHandlerRegistered)
    {
        free_irq(Dev->IrqNum,Dev);
        Dev->IrqHandlerRegistered=0;
    }

    if(Dev->MemHPI!=NULL)
    {
        iounmap(Dev->MemHPI);
        Dev->MemHPI=NULL;
    }

    if(Dev->MemGP!=NULL)
    {
        iounmap(Dev->MemGP);
        Dev->MemGP=NULL;
    }
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
irqreturn_t WRG3PCI_InterruptHandler(int irqno, void *dev_id)
#else
irqreturn_t WRG3PCI_InterruptHandler(int irqno, void *dev_id, struct pt_regs *regs)
#endif
{
 PWRG3 Dev;
 __u32 Request;
 unsigned long Flags;

    Dev=(PWRG3)dev_id;

    Request=ioread32(Dev->MemHPI+0x04);

    if(Request==0 || Request==0xFFFFFFFF)
    {
        return IRQ_NONE;
    }

    spin_lock_irqsave(&Dev->Interrupt_Lock,Flags);

    if(Dev->InterruptStarted!=0)
    {
        up(&Dev->Interrupt_Semaphore);
    }

    spin_unlock_irqrestore(&Dev->Interrupt_Lock,Flags);

    iowrite32(Request,Dev->MemHPI+0x04);

    return IRQ_HANDLED;
}

int WRG3PCI_ReadPIC(PWRG3 Dev,__u8 Address,__u8 *Value)
{
 __u16 RegistryAddress;
 int Retry;
 __u8 *Status;

    RegistryAddress=(Address & 0x3F) | 0x80;

    Status=Dev->MemGP+4;

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if(!(ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        ioread16(Dev->MemGP);

        Retry=MAX_PIC_RETRY;

        while(Retry--)
        {
            if(!(ioread16(Status)&0x0100))
            {
                break;
            }
        }

        if(Retry<=0)
        {
            return -EIO;
        }
    }

    iowrite16(RegistryAddress,Dev->MemGP);

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if(ioread16(Status)&0x0100)
        {
            break;
        }
    }

    if(Retry<=0)
    {
        return -EIO;
    }

    *Value=(__u8)(ioread16(Dev->MemGP) & 0xFF);

    return 0;
}

int WRG3PCI_WritePIC(PWRG3 Dev,__u8 Address,__u8 Value)
{
 __u16 RegistryAddress;
 __u8 *Status;
 int Retry;

    RegistryAddress=(Address&0x3F)|0x80|((Value&0x80)>>1);
    Status=Dev->MemGP+4;

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if(!(ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        ioread16(Dev->MemGP);

        Retry=MAX_PIC_RETRY;

        while(Retry--)
        {
            if(!(ioread16(Status)&0x0100))
            {
                break;
            }
        }

        if(Retry<=0)
        {
            return -EIO;
        }
    }

    iowrite16(RegistryAddress,Dev->MemGP);

    Retry=MAX_PIC_RETRY;

    while(Retry--)
    {
        if((ioread16(Status)&0x0100))
        {
            break;
        }
    }

    if(Retry<=0)
    {
        return -EIO;
    }

    iowrite16(Value&0x7F,Dev->MemGP);

    return 0;
}

static int WRG3PCI_Probe(struct pci_dev *dev, const struct pci_device_id *id)
{
 PWRG3 Dev;
 int Error;
 unsigned long ResourceStart;
 unsigned long ResourceLen;
 int i;
 int MemResIndex;

    Dev=kmalloc(sizeof(WRG3),GFP_KERNEL);

    if(Dev==NULL)
    {

        return -ENOMEM;
    }

    Error=pci_enable_device(dev);

    if(Error)
    {
        kfree(Dev);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to enable pci device (%d)\n",Error);
        return Error;
    }

    memset(Dev,0,sizeof(WRG3));

    Dev->PCIDevice=dev;
    Dev->InterfaceType=WRG3_PCI;
    Dev->DeviceIndex=-1;
    strcpy(Dev->Name,WRG3_DISPLAY_NAME);

    Dev->InterruptStarted=0;

    sema_init(&Dev->Interrupt_Semaphore,0);

    spin_lock_init(&Dev->Interrupt_Lock);

    WRG3_INIT_MUTEX(Dev->Ioctl_Lock);

    WRG3_LOCK(Dev->Ioctl_Lock);


    for(i=0,MemResIndex=0,Error=0;i<DEVICE_COUNT_RESOURCE && MemResIndex<3 && Error==0;i++)
    {
        if(pci_resource_flags(dev,i) & IORESOURCE_MEM)
        {
            ResourceStart=pci_resource_start(dev,i);
            ResourceLen=pci_resource_len(dev,i);

            switch(MemResIndex)
            {
                case 0:
                    Dev->MemHPI=(__u8*)ioremap(ResourceStart,ResourceLen);
                    if(Dev->MemHPI==NULL)
                    {
                        Error=-ENOMEM;
                    }
                    break;
                case 2:
                    Dev->MemGP=(__u8*)ioremap(ResourceStart,ResourceLen);
                    if(Dev->MemGP==NULL)
                    {
                        Error=-ENOMEM;
                    }
                    break;
                default:
                    break;
            }

            MemResIndex++;
        }
    }

    if(Dev->MemHPI==NULL || Dev->MemGP==NULL)
    {
        printk(KERN_ERR ": Unable to map memory resources.\n");
        Error=-ENOMEM;
    }

    if(Error!=0)
    {
        WRG3PCI_FreeResources(Dev);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

        kfree(Dev);
        return Error;
    }

    iowrite32(0xffffffff,Dev->MemHPI+0x04);
    iowrite32(0xffffffff,Dev->MemHPI+0x0C);



    Dev->IrqNum=dev->irq;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,22)
    Error=request_irq(dev->irq,WRG3PCI_InterruptHandler,IRQF_SHARED,WRG3_NAME"pci",Dev);
#else
    Error=request_irq(dev->irq,WRG3PCI_InterruptHandler,SA_INTERRUPT | SA_SHIRQ,WRG3_NAME"pci",Dev);
#endif

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to register interrupt handler (%d)\n",Error);
        WRG3PCI_FreeResources(Dev);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);
        kfree(Dev);
        return Error;
    }

    Dev->IrqHandlerRegistered=1;

    pci_set_drvdata(dev,Dev);

    pci_write_config_byte(dev, 0x44, 0x34);
    pci_write_config_byte(dev, 0x45, 0x00);
    pci_write_config_byte(dev, 0x46, 0x08);
    pci_write_config_byte(dev, 0x47, 0x00);
    pci_write_config_byte(dev, 0x47, 0x08);


    Error=RegisterWRG3Device(Dev);

    if(Error!=0)
    {
        WRG3PCI_FreeResources(Dev);
        pci_set_drvdata(dev,NULL);

        WRG3_UNLOCK(Dev->Ioctl_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);
        kfree(Dev);
        return Error;
    }

    WRG3PCI_WritePIC(Dev,0x05,0x00);

    Dev->Connected=1;

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return Error;
}

static void WRG3PCI_Remove(struct pci_dev *dev)
{
 PWRG3 Dev;

    Dev=(PWRG3)pci_get_drvdata(dev);

    if(Dev)
    {
        WRG3PCI_StopInterrupt(Dev);

        UnregisterWRG3Device(Dev);

        WRG3PCI_WritePIC(Dev,0x05,0x00);

        WRG3PCI_FreeResources(Dev);

        pci_disable_device(Dev->PCIDevice);
        pci_set_drvdata(Dev->PCIDevice,NULL);

        kfree(Dev);
    }
}

int WRG3USB_Configure(PWRG3 Dev,__u8 Alternate,__u8 ResetConfig);
static int WRG3USB_IsocUrbCompletionThread(void *Data);

static int WRG3USB_Probe(struct usb_interface *interface, const struct usb_device_id *id)
{
 struct usb_device *USBDevice;
 PWRG3 Dev;
 int Error;

    USBDevice=interface_to_usbdev(interface);

    Dev=kmalloc(sizeof(WRG3),GFP_KERNEL);

    if(Dev==NULL)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Out of memory.\n");
        return -ENOMEM;
    }

    memset(Dev,0,sizeof(WRG3));

    Dev->USBDevice=usb_get_dev(USBDevice);
    Dev->USBInterface=interface;
    Dev->InterfaceType=WRG3_USB;
    Dev->DeviceIndex=-1;
    strcpy(Dev->Name,WRG3_DISPLAY_NAME);

    WRG3_INIT_MUTEX(Dev->Ioctl_Lock);
    spin_lock_init(&Dev->SentBulkUrbs_SpinLock);

    WRG3_INIT_MUTEX(Dev->Streaming_Lock);
    WRG3_INIT_MUTEX(Dev->Streaming_Buffers_Lock);

    spin_lock_init(&Dev->Streaming_CompleteIsocUrbs_Lock);
    INIT_LIST_HEAD(&Dev->Streaming_CompleteIsocUrbs);
    sema_init(&Dev->Streaming_CompleteIsocUrbs_Semaphore,0);
    init_completion(&Dev->Streaming_Completion);
    init_rwsem(&Dev->Streaming_Semaphore);

    usb_set_intfdata(interface,Dev);

    Error=WRG3USB_Configure(Dev,0,1);

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Device configuration failed (%d)\n",Error);
        goto Error_Exit;
    }

    usb_clear_halt(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0));
    usb_clear_halt(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0));

    WRG3_LOCK(Dev->Ioctl_Lock);

    Error=RegisterWRG3Device(Dev);

    if(Error!=0)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        goto Error_Exit;
    }
    else
    {
        usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
        usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x03,USB_TYPE_VENDOR,0xA8,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
    }

    Dev->Connected=1;

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return 0;

    Error_Exit:

    usb_set_intfdata(interface,NULL);
    usb_put_dev(Dev->USBDevice);

    WRG3_DESTROY_MUTEX(Dev->Streaming_Buffers_Lock);
    WRG3_DESTROY_MUTEX(Dev->Streaming_Lock);
    WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

    kfree(Dev);

    return Error;
}

int WRG3USB_StopStreaming(PWRG3 Dev);

static void WRG3USB_Disconnect(struct usb_interface *interface)
{
 PWRG3 Dev;
 int i;

    Dev=usb_get_intfdata(interface);

    WRG3_LOCK(Dev->Ioctl_Lock);

    for(i=0;i<1000 && Dev->SentBulkUrbs!=0;i++)
    {
        msleep(50);
    }

    WRG3USB_StopStreaming(Dev);

    //WRG3USB_Configure(Dev,0);

    UnregisterWRG3Device(Dev);

    //usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(500));
    //usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x03,USB_TYPE_VENDOR,0x57,0,NULL,0,TO_JIFFIES(500));

    if(!Dev->Opened)
    {
        usb_set_intfdata(interface,NULL);

        usb_put_dev(Dev->USBDevice);

        WRG3_UNLOCK(Dev->Ioctl_Lock);

        WRG3_DESTROY_MUTEX(Dev->Streaming_Buffers_Lock);
        WRG3_DESTROY_MUTEX(Dev->Streaming_Lock);
        WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

        kfree(Dev);
    }
    else
    {
        Dev->Connected=0;

        WRG3_UNLOCK(Dev->Ioctl_Lock);
    }
}


static struct usb_driver WRG3USB_Driver=
{
    .name=WRG3_NAME,
    .probe=WRG3USB_Probe,
    .disconnect=WRG3USB_Disconnect,
    .id_table=WRG3USB_ID_Table,
};

static struct pci_driver WRG3PCI_Driver =
{
    .name=WRG3_NAME,
    .id_table=WRG3PCI_ID_Table,
    .probe=WRG3PCI_Probe,
    .remove=WRG3PCI_Remove,
};

static int __init WRG3_Init(void)
{
 int Error;
 dev_t FirstDev;



    FirstDev=MKDEV(WiNRADiO_MAJOR,WRG3_MINOR_BASE);

    Error=register_chrdev_region(FirstDev,MAX_WRG3_DEVICES,WRG3_NAME);

    if(Error<0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to register chrdev region (%d)\n",Error);
        return Error;
    }

    Error=class_register(&WRG3_Class);

    if(Error<0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to register class (%d)\n",Error);
        unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES);
        return Error;
    }

    Error=usb_register(&WRG3USB_Driver);

    if(Error!=0)
    {
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to register USB driver (%d)\n",Error);
    }
    else
    {
        Error=pci_register_driver(&WRG3PCI_Driver);

        if(Error!=0)
        {
            printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to register PCI driver (%d)\n",Error);
        }
    }

    if(Error!=0)
    {
        class_unregister(&WRG3_Class);
        unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES);
    }

    return Error;
}

static void __exit WRG3_Exit(void)
{
 dev_t FirstDev;

    usb_deregister(&WRG3USB_Driver);
    pci_unregister_driver(&WRG3PCI_Driver);

    class_unregister(&WRG3_Class);

    FirstDev=MKDEV(WiNRADiO_MAJOR,WRG3_MINOR_BASE);
    unregister_chrdev_region(FirstDev,MAX_WRG3_DEVICES);
}



int WRG3USB_Open(PWRG3 Dev)
{
    usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x03,USB_TYPE_VENDOR,0xA0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
    return 0;
}

int WRG3PCI_Open(PWRG3 Dev)
{
    return 0;
}

int WRG3_Open(struct inode *inode, struct file *filp)
{
 PWRG3 Dev;
 int Error;

    Dev=container_of(inode->i_cdev,WRG3,cdev);

    if(Dev==NULL)
    {
        return -ENOTTY;
    }

    if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
    {
        return -EINTR;
    }

    if(Dev->Opened)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EBUSY;
    }

    filp->private_data=Dev;

    switch(Dev->InterfaceType)
    {
        case WRG3_USB:
            Error=WRG3USB_Open(Dev);
            break;
        case WRG3_PCI:
            Error=WRG3PCI_Open(Dev);
            break;
        default:
            Error=-ENOTTY;
            break;
    }

    if(Error==0)
    {
        Dev->Opened=1;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return Error;
}

int WRG3PCI_Reset(PWRG3 Dev);

int WRG3_Release(struct inode *inode, struct file *filp)
{
 PWRG3 Dev;
 int i;

    Dev=(PWRG3)filp->private_data;

    if(Dev!=NULL)
    {
        WRG3_LOCK(Dev->Ioctl_Lock);

        Dev->Opened=0;

        if(Dev->InterfaceType==WRG3_USB)
        {
            for(i=0;i<1000 && Dev->SentBulkUrbs!=0;i++)
            {
                msleep(50);
            }

            WRG3USB_StopStreaming(Dev);

            if(Dev->Connected)
            {
                WRG3USB_Configure(Dev,0,1);

                usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x09,USB_TYPE_VENDOR,0,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
                usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0),0x03,USB_TYPE_VENDOR,0xA8,0,NULL,0,TO_JIFFIES(USB_TIMEOUT));
            }
            else
            {
                usb_set_intfdata(Dev->USBInterface,NULL);

                usb_put_dev(Dev->USBDevice);

                WRG3_UNLOCK(Dev->Ioctl_Lock);

                WRG3_DESTROY_MUTEX(Dev->Streaming_Buffers_Lock);
                WRG3_DESTROY_MUTEX(Dev->Streaming_Lock);
                WRG3_DESTROY_MUTEX(Dev->Ioctl_Lock);

                kfree(Dev);
            }
        }
        else
        {
            WRG3PCI_StopInterrupt(Dev);
            WRG3PCI_WritePIC(Dev,0x05,0x00);
            WRG3PCI_Reset(Dev);
        }

        WRG3_UNLOCK(Dev->Ioctl_Lock);
    }

    return 0;
}


int WRG3USB_ControlMessage_Send(PWRG3 Dev,unsigned long arg)
{
 int Error;
 void *Buff;
 u_int8_t Buffer[768];
 USB_CONTROL_MSG UsbMsg;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    Buff=Buffer;

    if(UsbMsg.Buffer!=NULL && UsbMsg.Size!=0)
    {
        if(!access_ok(VERIFY_READ,(void __user*)UsbMsg.Buffer,UsbMsg.Size))
        {
            return -EFAULT;
        }

        if(UsbMsg.Size>sizeof(Buffer))
        {
            Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

            if(Buff==NULL)
            {
                return -ENOMEM;
            }
        }

        if(copy_from_user(Buff,UsbMsg.Buffer,UsbMsg.Size))
        {
            if(Buff!=Buffer)
            {
                kfree(Buff);
            }
            return -EFAULT;
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USBDevice,0,
            UsbMsg.Request,USB_TYPE_VENDOR,
            UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
	usb_clear_halt(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0));
	usb_clear_halt(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0));

        Error=usb_control_msg(Dev->USBDevice,0,
                UsbMsg.Request,USB_TYPE_VENDOR,
                UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Error>=0)
    {
        if(Error==UsbMsg.Size)
        {
            Error=0;
        }
        else
        {
            Error=-EIO;
        }
    }

    if(Buff!=NULL && Buff!=Buffer)
    {
        kfree(Buff);
    }

    return Error;
}

int WRG3USB_ControlMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 USB_CONTROL_MSG UsbMsg;
 void *Buff;
 int Error;
 u_int8_t Buffer[768];


    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    Buff=Buffer;

    if(UsbMsg.Buffer!=NULL && UsbMsg.Size!=0)
    {
        if(!access_ok(VERIFY_WRITE,(void __user*)UsbMsg.Buffer,UsbMsg.Size))
        {
            return -EFAULT;
        }

        if(UsbMsg.Size>sizeof(Buffer))
        {
            Buff=kmalloc(UsbMsg.Size,GFP_KERNEL);

            if(Buff==NULL)
            {
                return -ENOMEM;
            }
        }
    }
    else
    {
        Buff=NULL;
        UsbMsg.Size=0;
    }

    Error=usb_control_msg(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0),
                UsbMsg.Request,USB_DIR_IN | USB_TYPE_VENDOR,
                UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0));
        usb_clear_halt(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0));

        Error=usb_control_msg(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0),
                    UsbMsg.Request,USB_DIR_IN | USB_TYPE_VENDOR,
                    UsbMsg.Value,UsbMsg.Index,Buff,UsbMsg.Size,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Buff!=NULL)
    {
        if(Error>0)
        {
            if(copy_to_user((void __user*)UsbMsg.Buffer,Buff,Error))//UsbMsg.Size))
            {
                Error=-EFAULT;
            }
        }

        if(Buff!=Buffer)
        {
            kfree(Buff);
        }
    }

    return Error;
}

int WRG3USB_ControlMessage(PWRG3 Dev,unsigned long arg)
{
 USB_CONTROL_MSG UsbMsg;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -ENOTTY;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&UsbMsg,(void __user*)arg,sizeof(UsbMsg)))
    {
        return -EFAULT;
    }

    Error=usb_control_msg(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0),
                UsbMsg.Request,USB_TYPE_VENDOR,
                UsbMsg.Value,UsbMsg.Index,NULL,0,TO_JIFFIES(UsbMsg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice,0));
        usb_clear_halt(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0));

        Error=usb_control_msg(Dev->USBDevice,usb_rcvctrlpipe(Dev->USBDevice,0),
                        UsbMsg.Request,USB_TYPE_VENDOR,
                        UsbMsg.Value,UsbMsg.Index,NULL,0,TO_JIFFIES(UsbMsg.Timeout));
    }

    if(Error>=0)
    {
        Error=0;
    }


    return Error;
}


int WRG3PCI_ReadPICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[448];
 __u8 ValuesBuff[448];
 __u8 *Address;
 __u8 *Value;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    Error=0;

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }

    if(!access_ok(VERIFY_READ,(void __user*)Reg.Addresses,Reg.ByteCount))
    {
        return -EFAULT;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)Reg.Values,Reg.ByteCount))
    {
        return -EFAULT;
    }


    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,(void __user*)Reg.Addresses,Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_ReadPIC(Dev,*Address,Value);

            if(Error!=0)
            {
                break;
            }
        }

        if(Error==0)
        {
            if(copy_to_user((void __user*)Reg.Values,Values,Reg.ByteCount))
            {
                Error=-EFAULT;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}

int WRG3PCI_WritePICBytes(PWRG3 Dev,unsigned long arg)
{
 int Error;
 PIC_REG Reg;
 __u16 i;
 __u8 *Addresses;
 __u8 *Values;
 __u8 AddressesBuff[448];
 __u8 ValuesBuff[448];
 __u8 *Address;
 __u8 *Value;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    Error=0;

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Reg,(void __user*)arg,sizeof(Reg)))
    {
        return -EFAULT;
    }

    if(Reg.ByteCount==0)
    {
        return 0;
    }

    if(!access_ok(VERIFY_READ,(void __user*)Reg.Addresses,Reg.ByteCount))
    {
        return -EFAULT;
    }

    if(!access_ok(VERIFY_READ,(void __user*)Reg.Values,Reg.ByteCount))
    {
        return -EFAULT;
    }


    if(Reg.ByteCount>sizeof(AddressesBuff))
    {
        Addresses=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Addresses==NULL)
        {
            return -ENOMEM;
        }

        Values=(__u8*)kmalloc(Reg.ByteCount,GFP_KERNEL);

        if(Values==NULL)
        {
            kfree(Addresses);
            return -ENOMEM;
        }
    }
    else
    {
        Addresses=AddressesBuff;
        Values=ValuesBuff;
    }

    if(copy_from_user(Addresses,(void __user*)Reg.Addresses,Reg.ByteCount)==0 &&
       copy_from_user(Values,(void __user*)Reg.Values,Reg.ByteCount)==0)
    {
        Address=Addresses;
        Value=Values;

        for(i=0;i<Reg.ByteCount;i++,Address++,Value++)
        {
            Error=WRG3PCI_WritePIC(Dev,*Address,*Value);

            if(Error!=0)
            {
                break;
            }
        }
    }
    else
    {
        Error=-EFAULT;
    }

    if(Addresses!=AddressesBuff)
    {
        kfree(Addresses);
        kfree(Values);
    }

    return Error;
}

int WRG3PCI_WritePICByte(PWRG3 Dev,unsigned long arg)
{
 __u8 PicReg[2];

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(__u16)))
    {
        return -EFAULT;
    }

    if(get_user((*((__u16*)PicReg)),(__u16 __user*)arg))
    {
        return -EFAULT;
    }

    return WRG3PCI_WritePIC(Dev,PicReg[0],PicReg[1]);
}

int WRG3PCI_ReadPICByte(PWRG3 Dev,unsigned long arg)
{
 __u8 PicReg[2];
 int Error;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(__u16)))
    {
        return -EFAULT;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(__u16)))
    {
        return -EFAULT;
    }

    if(get_user(*((__u16*)PicReg),(__u16 __user*)arg))
    {
        return -EFAULT;
    }

    Error=WRG3PCI_ReadPIC(Dev,PicReg[0],PicReg+1);

    if(Error==0)
    {
        if(put_user(*((__u16*)PicReg),(__u16 __user*)arg))
        {
            Error=-EFAULT;
        }
    }

    return Error;
}


#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
static void WRG3USB_IsoUrbComplete(struct urb *Urb)
#else
static void WRG3USB_IsoUrbComplete(struct urb *Urb,struct pt_regs *Regs)
#endif
{
 PWRG3 Dev;
 PISOC_URB IsocUrb;
 unsigned long Flags;

    IsocUrb=(PISOC_URB)Urb->context;
    Dev=(PWRG3)IsocUrb->Dev;

    spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

    if(Dev->Streaming_SentIsocUrbs>0)
    {
        list_add_tail(&IsocUrb->Urb->urb_list,&Dev->Streaming_CompleteIsocUrbs);

        Dev->Streaming_SentIsocUrbs--;
    }

    spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

    up(&Dev->Streaming_CompleteIsocUrbs_Semaphore);
}

static int WRG3USB_IsocUrbCompletionThread(void *Data)
{
 PWRG3 Dev;
 struct urb *Urb;
 struct list_head *ListEntry;
 unsigned long Flags;
 int i;
 struct usb_iso_packet_descriptor *Desc;
 PUSB_BUFFER Buffer;
 __u8 *Ptr;
 __u32 Size;
 int Error;


    Dev=(PWRG3)Data;

    up(&Dev->Streaming_Thread_Start_Semaphore);

    do
    {
        if(down_interruptible(&Dev->Streaming_CompleteIsocUrbs_Semaphore))
        {
            up(&Dev->Streaming_Thread_Stop_Semaphore);
            return 0;
        }


        if(Dev->Streaming_Thread_Stop)
        {
            break;
        }

        down_read(&Dev->Streaming_Semaphore);

        if(Dev->Streaming_Active)
        {
            spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

            if(list_empty(&Dev->Streaming_CompleteIsocUrbs))
            {
                Urb=NULL;
            }
            else
            {
                ListEntry=Dev->Streaming_CompleteIsocUrbs.next;

                Urb=list_entry(ListEntry,struct urb,urb_list);
                list_del(ListEntry);
                INIT_LIST_HEAD(ListEntry);
            }

            spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

            if(Urb!=NULL)
            {
                WRG3_LOCK(Dev->Streaming_Buffers_Lock);

                Desc=Urb->iso_frame_desc;

                Buffer=Dev->Streaming_CurrentBuffer;

                for(i=0;i<Urb->number_of_packets;i++,Desc++)
                {
                    if(Desc->status==0 && Desc->actual_length>0)
                    {
                        Ptr=((__u8*)Urb->transfer_buffer)+Desc->offset;
                        Size=Desc->actual_length;

                        if(Dev->Streaming_Shift>0)
                        {
                            if(Size>=Dev->Streaming_Shift)
                            {
                                Size-=Dev->Streaming_Shift;
                                Ptr+=Dev->Streaming_Shift;
                                Dev->Streaming_Shift=0;
                            }
                            else
                            {
                                Dev->Streaming_Shift-=Size;
                                Size=0;
                            }
                        }

                        while(Size>0)
                        {
                            if(Dev->Streaming_CurrentBuffer->Remain==0)
                            {
                                Buffer=Dev->Streaming_CurrentBuffer=Dev->Streaming_FirstBuffer;
                                Buffer->Used=0;
                                Buffer->Remain=Dev->Streaming_BufferSize;
                                Dev->Streaming_FirstBuffer=Dev->Streaming_FirstBuffer->Next;
                            }

                            if(Size>=Buffer->Remain)
                            {
                                memcpy(Buffer->Buffer+Buffer->Used,Ptr,Buffer->Remain);

                                Ptr+=Buffer->Remain;
                                Size-=Buffer->Remain;

                                Buffer->Remain=0;
                                Buffer->Used=Dev->Streaming_BufferSize;
                                Dev->Streaming_CurrentBuffer=Buffer=Buffer->Next;

                                complete(&Dev->Streaming_Completion);
                            }
                            else
                            {
                                memcpy(Buffer->Buffer+Buffer->Used,Ptr,Size);

                                Buffer->Used+=Size;
                                Buffer->Remain-=Size;
                                Size=0;
                            }
                        }
                    }
                }

                WRG3_UNLOCK(Dev->Streaming_Buffers_Lock);

                spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

                Dev->Streaming_SentIsocUrbs++;

                spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

                Error=usb_submit_urb(Urb,GFP_KERNEL);

                if(Error==-EPIPE)
                {
                    usb_clear_halt(Dev->USBDevice,Urb->pipe);
                    Error=usb_submit_urb(Urb,GFP_KERNEL);
                }

                if(Error!=0)
                {
                    spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

                    Dev->Streaming_SentIsocUrbs--;

                    spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);
                }
            }
        }

        up_read(&Dev->Streaming_Semaphore);
    }
    while(1);

    up(&Dev->Streaming_Thread_Stop_Semaphore);

    return 0;
}



int WRG3USB_Configure(PWRG3 Dev,__u8 Alternate,__u8 ResetConfig)
{
 //struct usb_host_interface *Alt;
 //struct usb_host_endpoint *EndPoint;
 __u32 i;
 int Error;

    WRG3USB_StopStreaming(Dev);

    if(ResetConfig)
    {
        usb_reset_configuration(Dev->USBDevice);
    }

    //usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice, 0),0x2C,USB_TYPE_VENDOR,5,0,NULL,0,USB_CTRL_SET_TIMEOUT);

    Error=usb_set_interface(Dev->USBDevice,Dev->USBInterface->altsetting[0].desc.bInterfaceNumber,Alternate);

    if(Error!=0)
    {
        return Error;
    }


    //Alt=Dev->USBInterface->cur_altsetting;
    //EndPoint=Alt->endpoint;

    Dev->NumberOfEndPoints=Dev->USBInterface->cur_altsetting->desc.bNumEndpoints;

    down_write(&Dev->Streaming_Semaphore);

    for(i=0;i<MAX_ENDPOINTS;i++)
    {
        if(i>0)
        {
            if(Dev->USBDevice->ep_in[i]!=NULL)
            {
                Error=usb_control_msg(Dev->USBDevice, usb_sndctrlpipe(Dev->USBDevice, 0),
                                USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                                USB_ENDPOINT_HALT, i|USB_DIR_IN, NULL, 0,
                                USB_CTRL_SET_TIMEOUT);
            }

            if(Dev->USBDevice->ep_out[i]!=NULL)
            {
                Error=usb_control_msg(Dev->USBDevice, usb_sndctrlpipe(Dev->USBDevice, 0),
                                USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                                USB_ENDPOINT_HALT, i, NULL, 0,
                                USB_CTRL_SET_TIMEOUT);
            }
        }
    }

    up_write(&Dev->Streaming_Semaphore);

    return 0;
}

int WRG3USB_GetEndPointInfo(PWRG3 Dev,unsigned long arg)
{
 ENDPOINTS EndPoints;
 __u32 i;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(EndPoints)))
    {
        return -EFAULT;
    }

    EndPoints.Count=Dev->NumberOfEndPoints;

    for(i=0;i<EndPoints.Count;i++)
    {
        EndPoints.EndPoint[i].Address=Dev->USBInterface->cur_altsetting->endpoint[i].desc.bEndpointAddress;
        EndPoints.EndPoint[i].Number=EndPoints.EndPoint[i].Address&USB_ENDPOINT_NUMBER_MASK;
    }

    if(copy_to_user((ENDPOINTS __user*)arg,&EndPoints,sizeof(EndPoints)))
    {
        return -EFAULT;
    }

    return 0;
}

int WRG3USB_StartStreaming(PWRG3 Dev,unsigned long arg)
{
 STREAM_INFO Info;
 __u32 i;
 PISOC_URB IsocUrb,PrevUrb;
 int Error;
 PUSB_BUFFER Buffer,Prev;
 unsigned long Flags;
 struct task_struct *thread;


    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(Dev->Streaming_Active)
    {
        return -EBUSY;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Info,(STREAM_INFO __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(Info.EndPoint>15 || Dev->USBDevice->ep_in[Info.EndPoint]==NULL || Info.NumberOfBuffers==0 || Info.BufferSize==0 ||
       Info.NumberOfUsbBuffers==0 || Info.UsbBufferSize==0)
    {
        return -EINVAL;
    }


    down_write(&Dev->Streaming_Semaphore);

    init_completion(&Dev->Streaming_Completion);

    Dev->Streaming_BufferSize=Info.BufferSize;

    Dev->Streaming_Shift=0;

    Prev=NULL;

    for(i=0;i<Info.NumberOfBuffers;i++)
    {
        Buffer=WRG3USB_AllocateBuffer(Info.BufferSize);

        if(Buffer==NULL)
        {
            WRG3USB_FreeStreamingBuffers(Dev);

            up_write(&Dev->Streaming_Semaphore);
            return -ENOMEM;
        }

        if(Prev!=NULL)
        {
            Prev->Next=Buffer;
        }
        else
        {
            Dev->Streaming_Buffers=Buffer;
        }

        Buffer->Next=Dev->Streaming_Buffers;
        Prev=Buffer;
    }

    Dev->Streaming_FirstBuffer=Dev->Streaming_Buffers;
    Dev->Streaming_CurrentBuffer=Dev->Streaming_Buffers;

    Dev->Streaming_Active=1;

    PrevUrb=NULL;

    for(i=0;i<Info.NumberOfUsbBuffers;i++)
    {
        Error=WRG3USB_AllocateIsocUrb(Dev,Info.EndPoint,Info.UsbBufferSize,&IsocUrb);

        if(Error!=0)
        {
            up_write(&Dev->Streaming_Semaphore);
            WRG3USB_StopStreaming(Dev);
            return Error;
        }

        if(PrevUrb!=NULL)
        {
            PrevUrb->Next=IsocUrb;
        }
        else
        {
            Dev->Streaming_IsocUrbs=IsocUrb;
        }

        IsocUrb->Next=Dev->Streaming_IsocUrbs;

        spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);
        Dev->Streaming_SentIsocUrbs++;
        spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

        Error=usb_submit_urb(IsocUrb->Urb,GFP_KERNEL);

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USBDevice,IsocUrb->Urb->pipe);
            Error=usb_submit_urb(IsocUrb->Urb,GFP_KERNEL);
        }

        if(Error!=0)
        {
            spin_lock_irqsave(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);
            Dev->Streaming_SentIsocUrbs--;
            spin_unlock_irqrestore(&Dev->Streaming_CompleteIsocUrbs_Lock,Flags);

            up_write(&Dev->Streaming_Semaphore);

            WRG3USB_StopStreaming(Dev);

            return Error;
        }
    }


    Dev->Streaming_Thread_Stop=0;

    //init_MUTEX_LOCKED(&Dev->Streaming_Thread_Start_Semaphore);
    //init_MUTEX_LOCKED(&Dev->Streaming_Thread_Stop_Semaphore);
    sema_init(&Dev->Streaming_Thread_Start_Semaphore,0);
    sema_init(&Dev->Streaming_Thread_Stop_Semaphore,0);

    thread=kthread_create(WRG3USB_IsocUrbCompletionThread,Dev,WRG3_NAME);

    if(IS_ERR(thread))
    {
        up_write(&Dev->Streaming_Semaphore);
        WRG3USB_StopStreaming(Dev);
        return PTR_ERR(thread);
    }

    Dev->Streaming_Thread=thread;

    wake_up_process(thread);

    down(&Dev->Streaming_Thread_Start_Semaphore);


    up_write(&Dev->Streaming_Semaphore);

    return 0;
}


int WRG3USB_StopStreaming(PWRG3 Dev)
{
 int i;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    down_write(&Dev->Streaming_Semaphore);

    if(Dev->Streaming_Active)
    {
        for(i=0;i<1000;i++)
        {
            if(Dev->Streaming_SentIsocUrbs==0)
            {
                break;
            }

            msleep(50);
        }

        WRG3USB_FreeIsocUrbs(Dev);

        Dev->Streaming_Active=0;

        complete_all(&Dev->Streaming_Completion);
    }

    up_write(&Dev->Streaming_Semaphore);

    if(Dev->Streaming_Thread!=NULL)
    {
        Dev->Streaming_Thread_Stop=1;
        up(&Dev->Streaming_CompleteIsocUrbs_Semaphore);

        down(&Dev->Streaming_Thread_Stop_Semaphore);

        Dev->Streaming_Thread=NULL;

        sema_init(&Dev->Streaming_CompleteIsocUrbs_Semaphore,0);
    }

    WRG3USB_FreeStreamingBuffers(Dev);

    return 0;
}

int WRG3USB_Streaming_Shift(PWRG3 Dev,unsigned long arg)
{
 PUSB_BUFFER Buffer,Prev;
 __u32 Shift;


    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(!Dev->Streaming_Active)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Shift)))
    {
        return -EFAULT;
    }

    if(get_user(Shift,(__u32 __user*)arg))
    {
        return -EFAULT;
    }

    WRG3_LOCK(Dev->Streaming_Buffers_Lock);

    if(Dev->Streaming_Buffers!=NULL)
    {
        while(Dev->Streaming_FirstBuffer->Used>0 && Dev->Streaming_FirstBuffer->Used<=Shift)
        {
            Shift-=Dev->Streaming_FirstBuffer->Used;
            Dev->Streaming_FirstBuffer->Used=0;
            Dev->Streaming_FirstBuffer->Remain=Dev->Streaming_BufferSize;
            Dev->Streaming_FirstBuffer=Dev->Streaming_FirstBuffer->Next;
        }

        if(Dev->Streaming_FirstBuffer->Used>0)
        {
            Dev->Streaming_Shift=0;

            if(Shift>0 && Dev->Streaming_FirstBuffer->Used>0)
            {
                Buffer=Dev->Streaming_FirstBuffer;
                Prev=NULL;

                do
                {
                    if(Prev!=NULL)
                    {
                        if(Buffer->Used<=Shift)
                        {
                            memcpy(Prev->Buffer+Prev->Used,Buffer->Buffer,Buffer->Used);
                            Prev->Used+=Buffer->Used;
                            Prev->Remain-=Buffer->Used;
                        }
                        else
                        {
                            memcpy(Prev->Buffer+Prev->Used,Buffer->Buffer,Shift);
                            Prev->Used+=Shift;
                            Prev->Remain-=Shift;
                        }
                    }

                    if(Buffer->Used<=Shift)
                    {
                        Buffer->Used=0;
                        Buffer->Remain=Dev->Streaming_BufferSize;
                    }
                    else
                    {
                        memcpy(Buffer->Buffer,Buffer->Buffer+Shift,Buffer->Used-Shift);
                        Buffer->Used-=Shift;
                        Buffer->Remain+=Shift;
                    }

                    Prev=Buffer;
                    Buffer=Buffer->Next;
                }
                while(Buffer->Used>0 && Buffer!=Dev->Streaming_FirstBuffer);
            }
        }
        else
        {
            Dev->Streaming_Shift=Shift;
        }

        Dev->Streaming_CurrentBuffer=Dev->Streaming_FirstBuffer;

        while(Dev->Streaming_CurrentBuffer->Remain==0)
        {
            Dev->Streaming_CurrentBuffer=Dev->Streaming_CurrentBuffer->Next;
        }
    }
    else
    {
        Dev->Streaming_Shift=Shift;
    }

    WRG3_UNLOCK(Dev->Streaming_Buffers_Lock);

    return 0;
}

int WRG3USB_GetStreamBuffer(PWRG3 Dev,unsigned long arg)
{
 STREAM_BUFFER Info;
 PUSB_BUFFER Buffer;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Info)))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(Info)))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }

    if(copy_from_user(&Info,(STREAM_BUFFER __user*)arg,sizeof(Info)))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }

    if(Info.BufferSize==0 || Info.Buffer==NULL)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)Info.Buffer,Info.BufferSize))
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EFAULT;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    do
    {
        down_read(&Dev->Streaming_Semaphore);

        if(Dev->Streaming_Active)
        {
            if(Dev->Streaming_BufferSize!=Info.BufferSize)
            {
                up_read(&Dev->Streaming_Semaphore);
                return -EINVAL;
            }

            if(WRG3_LOCK_INTERRUPTIBLE(Dev->Streaming_Buffers_Lock))
            {
                up_read(&Dev->Streaming_Semaphore);
                return -EINTR;
            }

            Buffer=Dev->Streaming_FirstBuffer;

            if(Buffer->Remain==0)
            {
                Dev->Streaming_FirstBuffer=Dev->Streaming_FirstBuffer->Next;

                Buffer->Remain=Dev->Streaming_BufferSize;
                Buffer->Used=0;

                if(copy_to_user((void __user*)Info.Buffer,Buffer->Buffer,Dev->Streaming_BufferSize))
                {
                    WRG3_UNLOCK(Dev->Streaming_Buffers_Lock);
                    up_read(&Dev->Streaming_Semaphore);
                    return -EFAULT;
                }

                Info.Filled=1;
                put_user(Info.Filled,(__u8 __user*)arg);

                WRG3_UNLOCK(Dev->Streaming_Buffers_Lock);
                up_read(&Dev->Streaming_Semaphore);
                break;
            }
            else
            {
                WRG3_UNLOCK(Dev->Streaming_Buffers_Lock);
            }
        }
        else
        {
            up_read(&Dev->Streaming_Semaphore);
            return -EINVAL;
        }

        up_read(&Dev->Streaming_Semaphore);

        if(Info.Wait)
        {
            if(wait_for_completion_interruptible(&Dev->Streaming_Completion))
            {
                return -EINTR;
            }
        }
        else
        {
            Info.Filled=0;
            put_user(Info.Filled,(__u8 __user*)arg);
            break;
        }
    }
    while(1);

    return 0;
}

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,19)
void WRG3xUSB_AsyncBulkUrbComplete(struct urb *Urb)
#else
void WRG3xUSB_AsyncBulkUrbComplete(struct urb *Urb,struct pt_regs *regs)
#endif
{
 PWRG3 Dev;
 unsigned long Flags;

    Dev=(PWRG3)Urb->context;

    spin_lock_irqsave(&Dev->SentBulkUrbs_SpinLock,Flags);

    Dev->SentBulkUrbs--;

    spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);

    kfree(Urb->transfer_buffer);
    usb_free_urb(Urb);
}

int WRG3USB_BulkMessage_Send_Async(PWRG3 Dev,unsigned long arg)
{
 USB_BULK_MESSAGE Msg;
 __u8 *Buff;
 int Error;
 unsigned long Flags;
 int i;
 struct urb *Urb;


    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Msg,(USB_BULK_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USBDevice->ep_out[Msg.EndPoint]==NULL || Msg.Buffer==NULL || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)Msg.Buffer,Msg.BufferSize))
    {
        return -EFAULT;
    }

    Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

    if(Buff==NULL)
    {
        return -ENOMEM;
    }

    if(copy_from_user(Buff,(__u8 __user*)Msg.Buffer,Msg.BufferSize))
    {
        Error=-EFAULT;
    }
    else
    {
        for(i=0;1;i++)
        {
            if(i>1000)
            {
                kfree(Buff);
                return -EBUSY;
            }

            spin_lock_irqsave(&Dev->SentBulkUrbs_SpinLock,Flags);

            if(Dev->SentBulkUrbs<MAX_BULK_URBS)
            {
                Dev->SentBulkUrbs++;
                spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);
                break;
            }

            spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);

            msleep(10);
        }


        Urb=usb_alloc_urb(0,GFP_KERNEL);

        if(Urb!=NULL)
        {
            usb_fill_bulk_urb(Urb,Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint),Buff,
               Msg.BufferSize,WRG3xUSB_AsyncBulkUrbComplete,Dev);

            Error=usb_submit_urb(Urb,GFP_KERNEL);

            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint));

                Error=usb_submit_urb(Urb,GFP_KERNEL);
            }

            if(Error!=0)
            {
                usb_free_urb(Urb);
            }
        }
        else
        {
            Error=-ENOMEM;
        }

        if(Error!=0)
        {
            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint));
            }

            spin_lock_irqsave(&Dev->SentBulkUrbs_SpinLock,Flags);

            Dev->SentBulkUrbs--;

            spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);

            kfree(Buff);
        }
    }

    return Error;
}

int WRG3USB_BulkMessage_Send(PWRG3 Dev,unsigned long arg)
{
 USB_BULK_MESSAGE Msg;
 __u8 Buffer[900];
 __u8 *Buff;
 int Error;
 unsigned long Flags;
 int i;


    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Msg,(USB_BULK_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(Msg.EndPoint>15 || Dev->USBDevice->ep_out[Msg.EndPoint]==NULL || Msg.Buffer==NULL || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)Msg.Buffer,Msg.BufferSize))
    {
        return -EFAULT;
    }

    if(Msg.BufferSize>sizeof(Buffer))
    {
        Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Buffer;
    }

    if(copy_from_user(Buff,(__u8 __user*)Msg.Buffer,Msg.BufferSize))
    {
        Error=-EFAULT;
    }
    else
    {
        for(i=0;1;i++)
        {
            if(i>100)
            {
                if(Buff!=Buffer)
                {
                    kfree(Buff);
                }
                return -EBUSY;
            }

            spin_lock_irqsave(&Dev->SentBulkUrbs_SpinLock,Flags);

            if(Dev->SentBulkUrbs<MAX_BULK_URBS)
            {
                Dev->SentBulkUrbs++;
                spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);
                break;
            }

            spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);

            msleep(10);
        }


        Error=usb_bulk_msg(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint),
              Buff,Msg.BufferSize,&Msg.ActualSize,TO_JIFFIES(Msg.Timeout));

        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint));

            Error=usb_bulk_msg(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint),
                    Buff,Msg.BufferSize,&Msg.ActualSize,TO_JIFFIES(Msg.Timeout));
        }

        spin_lock_irqsave(&Dev->SentBulkUrbs_SpinLock,Flags);

        Dev->SentBulkUrbs--;

        spin_unlock_irqrestore(&Dev->SentBulkUrbs_SpinLock,Flags);

        if(Error==0)
        {
            if(copy_to_user((USB_BULK_MESSAGE __user*)arg,&Msg,sizeof(Msg)))
            {
                Error=-EFAULT;
            }
        }
        else
        {
            if(Error==-EPIPE)
            {
                usb_clear_halt(Dev->USBDevice,usb_sndbulkpipe(Dev->USBDevice,Msg.EndPoint));
            }
        }
    }

    if(Buff!=Buffer)
    {
        kfree(Buff);
    }

    return Error;
}

int WRG3USB_BulkMessage_Receive(PWRG3 Dev,unsigned long arg)
{
 USB_BULK_MESSAGE Msg;
 __u8 Buffer[900];
 __u8 *Buff;
 int Error;

    if(Dev->InterfaceType!=WRG3_USB)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Msg,(USB_BULK_MESSAGE __user*)arg,sizeof(Msg)))
    {
        return -EFAULT;
    }


    if(Msg.EndPoint>15 || Dev->USBDevice->ep_in[Msg.EndPoint]==NULL || Msg.Buffer==NULL || Msg.BufferSize==0)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_WRITE,(void __user*)Msg.Buffer,Msg.BufferSize))
    {
        return -EFAULT;
    }

    if(Msg.BufferSize>sizeof(Buffer))
    {
        Buff=(__u8*)kmalloc(Msg.BufferSize,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Buffer;
    }

    Error=usb_bulk_msg(Dev->USBDevice,usb_rcvbulkpipe(Dev->USBDevice,Msg.EndPoint),
          Buff,Msg.BufferSize,&Msg.ActualSize,TO_JIFFIES(Msg.Timeout));

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USBDevice,usb_rcvbulkpipe(Dev->USBDevice,Msg.EndPoint));

        Error=usb_bulk_msg(Dev->USBDevice,usb_rcvbulkpipe(Dev->USBDevice,Msg.EndPoint),
                Buff,Msg.BufferSize,&Msg.ActualSize,TO_JIFFIES(Msg.Timeout));
    }

    if(Error==0)
    {
        if(copy_to_user((USB_BULK_MESSAGE __user*)arg,&Msg,sizeof(Msg)))
        {
            Error=-EFAULT;
        }

        if(copy_to_user((__u8 __user*)Msg.Buffer,Buff,Msg.BufferSize))
        {
            Error=-EFAULT;
        }
    }
    else
    {
        if(Error==-EPIPE)
        {
            usb_clear_halt(Dev->USBDevice,usb_rcvbulkpipe(Dev->USBDevice,Msg.EndPoint));
        }
    }

    if(Buff!=Buffer)
    {
        kfree(Buff);
    }

    return Error;
}

int WRG3PCI_Write(PWRG3 Dev,unsigned long arg)
{
 READ_WRITE_INFO Info;
 __u16 Buffer[448];
 __u16 *Buff,*Ptr;
 __u32 Size;


    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Info,(READ_WRITE_INFO __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(Info.Buffer==NULL || Info.Count==0)
    {
        return -EINVAL;
    }

    Size=Info.Count*sizeof(__u16);

    if(!access_ok(VERIFY_READ,(void __user*)arg,Size))
    {
        return -EINVAL;
    }

    if(Size>sizeof(Buffer))
    {
        Buff=(__u16*)kmalloc(Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Buffer;
    }


    if(copy_from_user(Buff,(__u16 __user*)Info.Buffer,Size))
    {
        if(Buff!=Buffer)
        {
            kfree(Buff);
        }
        return -EFAULT;
    }

    Ptr=Buff;

    iowrite32(Info.Address>>16,Dev->MemGP+0x18);
    iowrite32(Info.Address&0xFFFF,Dev->MemGP+0x18);

    while(Info.Count--)
    {
        iowrite32(*Ptr++,Dev->MemGP+0x08);
    }

    if(Buff!=Buffer)
    {
        kfree(Buff);
    }

    return 0;
}

int WRG3PCI_Read(PWRG3 Dev,unsigned long arg)
{
 READ_WRITE_INFO Info;
 __u16 Buffer[448];
 __u16 *Buff;
 __u32 Size;

    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(copy_from_user(&Info,(READ_WRITE_INFO __user*)arg,sizeof(Info)))
    {
        return -EFAULT;
    }

    if(Info.Buffer==NULL || Info.Count==0)
    {
        return -EINVAL;
    }

    Size=Info.Count*sizeof(__u16);

    if(!access_ok(VERIFY_WRITE,(void __user*)arg,Size))
    {
        return -EINVAL;
    }

    if(Size>sizeof(Buffer))
    {
        Buff=(__u16*)kmalloc(Size,GFP_KERNEL);

        if(Buff==NULL)
        {
            return -ENOMEM;
        }
    }
    else
    {
        Buff=Buffer;
    }


    iowrite32(Info.Address>>16,Dev->MemGP+0x18);
    iowrite32(Info.Address&0xFFFF,Dev->MemGP+0x18);

    if(Info.Address&0x40000000)
    {
        ioread16(Dev->MemGP+0x08);
    }

    ioread16_rep(Dev->MemGP+0x08,Buff,Info.Count);

    if(copy_to_user((__u16 __user*)Info.Buffer,Buff,Size))
    {
        if(Buff!=Buffer)
        {
            kfree(Buff);
        }
        return -EFAULT;
    }

    if(Buff!=Buffer)
    {
        kfree(Buff);
    }

    return 0;
}

int WRG3PCI_Reset(PWRG3 Dev)
{
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        return -EINVAL;
    }

    pci_write_config_byte(Dev->PCIDevice, 0x47, 0x00);
    pci_write_config_byte(Dev->PCIDevice, 0x47, 0x08);

    return 0;
}

int WRG3PCI_CatchInterrupt(PWRG3 Dev)
{
    if(Dev->InterfaceType!=WRG3_PCI)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    if(Dev->InterruptStarted==0)
    {
        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return -EINVAL;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    if(down_interruptible(&Dev->Interrupt_Semaphore))
    {
        return -EINTR;
    }

    return Dev->InterruptStarted?0:-EINVAL;
}

int WRG3PCI_StartInterrupt(PWRG3 Dev)
{
 unsigned long Flags;

    spin_lock_irqsave(&Dev->Interrupt_Lock,Flags);

    if(Dev->InterruptStarted==0)
    {
        Dev->InterruptStarted=1;
        sema_init(&Dev->Interrupt_Semaphore,0);
        iowrite32(0x8C000000,Dev->MemHPI+0x08);
    }

    spin_unlock_irqrestore(&Dev->Interrupt_Lock,Flags);

    return 0;
}

int WRG3PCI_StopInterrupt(PWRG3 Dev)
{
 int i;
 unsigned long Flags;

    spin_lock_irqsave(&Dev->Interrupt_Lock,Flags);

    if(Dev->InterruptStarted!=0)
    {
        Dev->InterruptStarted=0;
        iowrite32(0xFFFFFFFF,Dev->MemHPI+0x04);
        iowrite32(0xFFFFFFFF,Dev->MemHPI+0x0C);
    }

    spin_unlock_irqrestore(&Dev->Interrupt_Lock,Flags);

    for(i=0;i<10;i++)
    {
        up(&Dev->Interrupt_Semaphore);
    }

    return 0;
}

int WRG3USB_ResetEp(PWRG3 Dev,unsigned long arg)
{
/*
 __u8 EndPoint;

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(__u8)))
    {
        return -EFAULT;
    }

    if(get_user(EndPoint,(__u8 __user*)arg))
    {
        return -EFAULT;
    }
*/
    return 0;
}

int WRG3USB_ClearHalt(PWRG3 Dev,unsigned long arg)
{
 int Error;
 //int PipeOut;
 __u8 In;
 __u8 EndPoint;

    if(!access_ok(VERIFY_READ,(void __user*)arg,sizeof(__u8)))
    {
        return -EFAULT;
    }

    if(get_user(EndPoint,(__u8 __user*)arg))
    {
        return -EFAULT;
    }

    if(EndPoint&0x80)
    {
        EndPoint=(EndPoint&0x7F);
        In=USB_DIR_IN;
        //PipeOut=0;
    }
    else
    {
        //PipeOut=1;
        In=0;
    }


    Error=usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice, 0),
                          USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                          USB_ENDPOINT_HALT, EndPoint|In, NULL, 0,
                          USB_CTRL_SET_TIMEOUT);

    if(Error==-EPIPE)
    {
        usb_clear_halt(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice, 0));

        Error=usb_control_msg(Dev->USBDevice,usb_sndctrlpipe(Dev->USBDevice, 0),
                             USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT,
                             USB_ENDPOINT_HALT, EndPoint|In, NULL, 0,
                             USB_CTRL_SET_TIMEOUT);
    }

    if(Error<0)
    {
        return Error;
    }

    return 0;
}


#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,37)
long WRG3_Ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
#else
int WRG3_Ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
#endif
{
 PWRG3 Dev;
 int Error;
 u_int8_t Value8;

    Dev=(PWRG3)filp->private_data;

    if(Dev==NULL)
    {
        return -EIO;
    }


    if(WRG3_LOCK_INTERRUPTIBLE(Dev->Ioctl_Lock))
    {
        return -EINTR;
    }

    if(cmd==WRG3_IOCTL_IS_CONNECTED)
    {
        if(access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(__u8)))
        {
            if(Dev->Connected)
            {
                Value8=1;
            }
            else
            {
                Value8=0;
            }

            put_user(Value8,(__u8 __user*)arg);

            Error=0;
        }
        else
        {
            Error=-EFAULT;
        }


        WRG3_UNLOCK(Dev->Ioctl_Lock);
        return Error;
    }
    else
    {
        if(!Dev->Connected)
        {
            WRG3_UNLOCK(Dev->Ioctl_Lock);
            return -EIO;
        }
    }


    switch(cmd)
    {
        case WRG3_IOCTL_PCI_READ_PIC_BYTE:
            Error=WRG3PCI_ReadPICByte(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE_PIC_BYTE:
            Error=WRG3PCI_WritePICByte(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_READ_PIC_BYTES:
            Error=WRG3PCI_ReadPICBytes(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE_PIC_BYTES:
            Error=WRG3PCI_WritePICBytes(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_SEND_ASYNC:
            Error=WRG3USB_BulkMessage_Send_Async(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_SEND:
            Error=WRG3USB_BulkMessage_Send(Dev,arg);
            break;
        case WRG3_IOCTL_USB_BULK_MESSAGE_RECEIVE:
            Error=WRG3USB_BulkMessage_Receive(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_WRITE:
            Error=WRG3PCI_Write(Dev,arg);
            break;
        case WRG3_IOCTL_PCI_READ:
            Error=WRG3PCI_Read(Dev,arg);
            break;
        case WRG3_IOCTL_USB_GET_STREAM_BUFFER:
            return WRG3USB_GetStreamBuffer(Dev,arg);
        case WRG3_IOCTL_PCI_CATCH_INTERRUPT:
            return WRG3PCI_CatchInterrupt(Dev);
        case WRG3_IOCTL_USB_CONTROL_MESSAGE:
            Error=WRG3USB_ControlMessage(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONTROL_MESSAGE_SEND:
            Error=WRG3USB_ControlMessage_Send(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONTROL_MESSAGE_RECEIVE:
            Error=WRG3USB_ControlMessage_Receive(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CONFIGURE:
            if(access_ok(VERIFY_READ,(void __user*)arg,sizeof(__u8)))
            {
                if(!get_user(Value8,(__u8 __user*)arg))
                {
                    Error=WRG3USB_Configure(Dev,Value8,0);
                }
                else
                {
                    Error=-EFAULT;
                }
            }
            else
            {
                Error=-EFAULT;
            }
            break;
        case WRG3_IOCTL_USB_GET_ENDPOINTINFO:
            Error=WRG3USB_GetEndPointInfo(Dev,arg);
            break;
        case WRG3_IOCTL_USB_START_STREAMING:
            Error=WRG3USB_StartStreaming(Dev,arg);
            break;
        case WRG3_IOCTL_USB_STOP_STREAMING:
            Error=WRG3USB_StopStreaming(Dev);
            break;
        case WRG3_IOCTL_PCI_START_INTERRUPT:
            Error=WRG3PCI_StartInterrupt(Dev);
            break;
        case WRG3_IOCTL_PCI_STOP_INTERRUPT:
            Error=WRG3PCI_StopInterrupt(Dev);
            break;
        case WRG3_IOCTL_GET_INTERFACE:
            if(access_ok(VERIFY_WRITE,(void __user*)arg,sizeof(__u8)))
            {
                put_user(Dev->InterfaceType,(__u8 __user*)arg);

                Error=0;
            }
            else
            {
                Error=-EFAULT;
            }
            break;
        case WRG3_IOCTL_PCI_RESET:
            Error=WRG3PCI_Reset(Dev);
            break;
        case WRG3_IOCTL_USB_RESET_EP:
            Error=WRG3USB_ResetEp(Dev,arg);
            break;
        case WRG3_IOCTL_USB_CLEAR_HALT:
            Error=WRG3USB_ClearHalt(Dev,arg);
            break;
        case WRG3_IOCTL_USB_SHIFT_STREAM:
            Error=WRG3USB_Streaming_Shift(Dev,arg);
            break;
        default:
            Error=-EINVAL;
            break;
    }

    WRG3_UNLOCK(Dev->Ioctl_Lock);

    return Error;
}

static void WRG3_ReleaseCallback(struct device *Devive)
{
}

int RegisterWRG3Device(PWRG3 Dev)
{
 int i;
 int DeviceIndex;
 int Minor;
 dev_t DevNo;
 int Error;

    DeviceIndex=-1;


    if(WRG3_LOCK_INTERRUPTIBLE(WRG3_Device_Lock))
    {
        return -EINTR;
    }

    if(Dev->DeviceIndex>=0)
    {
        //device is already registered

        WRG3_UNLOCK(WRG3_Device_Lock);
        return 0;
    }

    for(i=0;i<MAX_WRG3_DEVICES;i++)
    {
        if(WRG3_Devices[i]==NULL)
        {
            DeviceIndex=i;
            break;
        }
    }

    if(DeviceIndex<0)
    {
        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Maximum number of supported devices [%d] by driver excess.\n",MAX_WRG3_DEVICES);
        return -ENFILE;
    }

    Minor=WRG3_MINOR_BASE+DeviceIndex;

    DevNo=MKDEV(WiNRADiO_MAJOR,Minor);

    Dev->Minor=Minor;
    Dev->DeviceIndex=DeviceIndex;

    memset(&Dev->Device,0,sizeof(Dev->Device));

    Dev->Device.class=&WRG3_Class;
    Dev->Device.devt=MKDEV(WiNRADiO_MAJOR,Minor);
    Dev->Device.release=WRG3_ReleaseCallback;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,26)
    dev_set_name(&Dev->Device,WRG3_NAME"%d",DeviceIndex);
#else
    snprintf(Dev->Device.class_id,BUS_ID_SIZE,WRG3_NAME"%d",DeviceIndex);
#endif

    Error=device_register(&Dev->Device);

    if(Error<0)
    {
        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Failed to register class device (%d).\n",Error);
        return Error;
    }

    cdev_init(&Dev->cdev,&WRG3_File_Operations);
    Dev->cdev.owner=THIS_MODULE;
    Dev->cdev.ops=&WRG3_File_Operations;
    Error=cdev_add(&Dev->cdev,DevNo,1);

    if(Error<0)
    {
        device_unregister(&Dev->Device);

        WRG3_UNLOCK(WRG3_Device_Lock);
        printk(KERN_ERR WRG3_DISPLAY_NAME ": Unable to add character device (%d).\n",Error);
        return Error;
    }

    WRG3_Devices[DeviceIndex]=Dev;

    WRG3_UNLOCK(WRG3_Device_Lock);

    return 0;
}

int UnregisterWRG3Device(PWRG3 Dev)
{
    if(WRG3_LOCK_INTERRUPTIBLE(WRG3_Device_Lock))
    {
        return -EINTR;
    }

    if(Dev->DeviceIndex>=0 && Dev->DeviceIndex<MAX_WRG3_DEVICES)
    {
        if(WRG3_Devices[Dev->DeviceIndex]==Dev)
        {
            WRG3_Devices[Dev->DeviceIndex]=NULL;
            cdev_del(&Dev->cdev);

            device_unregister(&Dev->Device);
        }
    }

    WRG3_UNLOCK(WRG3_Device_Lock);

    return 0;
}

module_init(WRG3_Init);
module_exit(WRG3_Exit);

MODULE_LICENSE("GPL");

//EXPORT_SYMBOL_GPL();
